import math
from typing import List, Dict, Any, Optional, Union

import numpy as np
import numpy.typing as npt

from ataims.structure import Structure
from ataims.output_aims import OutputAims
from ataims.output_exciting import OutputExciting
# from ..common import user_msg_box


def show_element(element, show: bool = True) -> None:
    element.style.display = '' if show else 'none'


# def on_error(message: str, url: str, line: int, col: int, error: Exception) -> None:
    # user_msg_box.set_error(f"{message}\n At {line}:{col} of {url}")


def get_output_instance(files_data: List[Dict[str, Any]]) -> Union[OutputAims, OutputExciting, None]:
    output_aims = OutputAims()
    output_exciting = OutputExciting()  # untested
    output_exciting.parse_files(files_data)
    if output_exciting.is_filled_up():
        return output_exciting
    output = output_aims.parse_files(files_data)
    if output.is_filled_up():
        return output
    return None


def geo_text_from_structure(structure: Structure) -> str:
    text = ''
    lattice_vectors = structure.lat_vectors
    if lattice_vectors:
        for v in lattice_vectors:
            text += f"lattice_vector {' '.join(map(str, v))}\n"
    for atom in structure.atoms:
        text += f"atom {' '.join(map(str, atom['position']))} {atom['species']}\n"
    return text


def line_array_to_text(lines: List[str]) -> str:
    return '\n'.join(lines) + '\n'


def get_parallelepiped_volume(vectors: List[List[float]]) -> float:
    #  volume = |a · (b × c)|, where a, b, and c are the three vectors defining the parallelepiped.
    v0 = np.array(vectors[0])
    v1 = np.array(vectors[1])
    v2 = np.array(vectors[2])
    return abs(np.dot(np.cross(v0, v1), v2))


def gauss(broad: float, x: float, x0: float) -> float:
    return 1.0 / (broad * math.sqrt(2 * math.pi)) * math.exp(-0.5 * (x - x0)**2 / broad**2)


def get_html_rows(quantities: Dict[str, Any], values_as_numbers: bool = True) -> str:
    html = ''
    for label, quantity in quantities.items():
        if label == 'download-link':
            html += f'<tr> <td><a class="f-geometry-download-link" download="geometry.in" href="{quantity}" >Download final geometry </a></td></tr>'
        else:
            value = f"{quantity:.5f}" if values_as_numbers else quantity
            html += f'<tr> <td>{label}</td> <td>{value}</td> </tr>'
    return html


# Cool-warm colormap from MPL
COLORS = np.array([
    [0.2481, 0.326, 0.7777], [0.3992, 0.5285, 0.9285], [0.5652, 0.6994, 0.9966],
    [0.729, 0.8175, 0.9732], [0.8674, 0.8644, 0.8626], [0.9582, 0.7712, 0.6803],
    [0.9594, 0.6103, 0.4894], [0.8771, 0.3946, 0.3117], [0.7233, 0.0689, 0.163]
])
COLOR_SPACE = len(COLORS) - 1


def rgb(r: float, g: float, b: float) -> str:
    return f"rgb({int(r)},{int(g)},{int(b)})"


def color(x: float) -> Optional[str]:
    if x < 0 or x > 1:
        return None
    if x == 0:
        return rgb(*(COLORS[0] * 255).astype(int))
    if x == 1:
        return rgb(*(COLORS[-1] * 255).astype(int))
    lo = int(x * COLOR_SPACE)
    hi = math.ceil(x * COLOR_SPACE)
    c = ((hi - x * COLOR_SPACE) * COLORS[lo] + (x * COLOR_SPACE - lo) * COLORS[hi]) * 255
    return rgb(*c.astype(int))


def array_to_colors(x: npt.NDArray[np.float64], around_zero: bool = False) -> List[str]:
    abs_max = np.max(np.abs(x))
    hi = abs_max if around_zero else np.max(x)
    lo = -abs_max if around_zero else np.min(x)
    print(lo, hi)
    return [color((i - lo) / (hi - lo)) for i in x]


def deep_merge(target: Dict[str, Any], update: Dict[str, Any]) -> Dict[str, Any]:
    for key, value in update.items():
        if isinstance(value, (str, int, float, bool)) or isinstance(value, list):
            target[key] = value
        elif isinstance(value, dict):
            target[key] = deep_merge(target.get(key, {}), value)
    return target


def transpose(arr: List[List[Any]]) -> List[List[Any]]:
    return list(map(list, zip(*arr)))

# not used for current purposes
# async def get_structure_info(structure: Structure) -> Optional[Dict[str, Any]]:
#     """
#     Gets structure information from the backend.

#     Args:
#         structure (Structure): A structure to get info on.

#     Returns:
#         Optional[Dict[str, Any]]: A dictionary with structure information, or None if no atoms.

#     Raises:
#         Exception: If there's an HTTP error during the request.
#     """
#     if not structure.atoms:
#         return None

#     send_structure = {
#         "cell": structure.lat_vectors,
#         "positions": structure.atoms,
#         "symThresh": Conf.settings.symmetry_threshold  # Assuming Conf is imported and defined
#     }

#     async with aiohttp.ClientSession() as session:
#         async with session.post(
#             '/update-structure-info',
#             headers={'Content-Type': 'application/json;charset=utf-8'},
#             json=send_structure
#         ) as response:
#             if not response.ok:
#                 UserMsgBox.set_error('Unknown server ERROR')  # Assuming UserMsgBox is imported and defined
#                 raise Exception(f"HTTP error: {response.status}")
            
#             structure_data = await response.json()

#     return structure_data.get('structureInfo')


async def get_structure_info(structure: Structure) -> Optional[Dict[str, Any]]:
    pass



"""
@author Iker Hurtado, Sebastian Kokott (Original)

@overview This is a basic math library adapted to use NumPy and Pandas.
It maintains the functionality of the original JavaScript implementation.
"""
import numpy as np
import pandas as pd
from typing import List, Union, Callable, Optional


def add(p1: List[float], p2: List[float]) -> List[float]:
    """Adds two vectors and returns a new one"""
    return np.add(p1, p2).tolist()


def add_up(p1: List[float], p2: List[float]) -> List[float]:
    """Adds up a vector (the second parameter) to first one and returns it"""
    p1 += np.array(p2)
    return p1.tolist()


def subtract(p2: List[float], p1: List[float]) -> List[float]:
    """Subtracts the second vector from the first one and returns a new one"""
    return np.subtract(p2, p1).tolist()


def multiply_scalar(vector: List[float], scalar: float) -> List[float]:
    """Multiplies a vector by a scalar number and returns a new vector"""
    return (np.array(vector) * scalar).tolist()


def divide_scalar(vector: List[float], scalar: float) -> List[float]:
    """Divides a vector by a scalar number and returns a new vector"""
    return (np.array(vector) / scalar).tolist()


def get_distance(p1: List[float], p2: List[float]) -> float:
    """Returns the distance between two points"""
    return np.linalg.norm(np.subtract(p1, p2))


def norm(v: List[float]) -> float:
    """Returns vector length"""
    return np.linalg.norm(v)


def cdot(x: List[float], y: List[float]) -> float:
    """Returns dot product of two 3D vectors"""
    return np.dot(x, y)


def cross(a: List[float], b: List[float]) -> List[float]:
    """Returns cross product of two 3D vectors"""
    return np.cross(a, b).tolist()


def angle(x: List[float], y: List[float]) -> float:
    """Returns angle between two 3D vectors"""
    return np.arccos(np.clip(np.dot(x, y) / (norm(x) * norm(y)), -1.0, 1.0))


def normalize(v: List[float]) -> List[float]:
    """Normalizes 3D vector v"""
    return (np.array(v) / norm(v)).tolist()


def get_angle(p1: List[float], p2: List[float], p3: List[float]) -> float:
    """Returns the angle made up of three points"""
    p12v = subtract(p1, p2)
    p32v = subtract(p3, p2)
    return np.degrees(angle(p12v, p32v))


def get_torsion_angle(p0: List[float], p1: List[float], p2: List[float], p3: List[float]) -> float:
    """Returns the torsion angle comprised of four points using praxeolitic formula"""
    b0 = np.subtract(p0, p1)
    b1 = np.subtract(p2, p1)
    b2 = np.subtract(p3, p2)
    b1 = b1 / np.linalg.norm(b1)
    v = b0 - np.dot(b0, b1) * b1
    w = b2 - np.dot(b2, b1) * b1
    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)
    return np.degrees(np.arctan2(y, x))


def minmax(mat: List[List[float]], axis: int, op: Callable[[np.ndarray], float]) -> List[float]:
    """Determines minimum or maximum of array along certain axis."""
    df = pd.DataFrame(mat)
    return df.apply(op, axis=axis).tolist()


def matrix_dot(A: List[List[float]], B: List[List[float]], C: Optional[List[List[float]]] = None) -> List[List[float]]:
    """Multiplies (dot) two matrices. C can be pre-allocated for performance."""
    if C is None:
        return np.dot(A, B).tolist()
    np.dot(A, B, out=C)
    return C


def tensor_dot(A: List[List[float]], B: List[float]) -> List[List[float]]:
    """Not yet full tensor dot. Accepts vector B and matrix A."""
    return np.tensordot(A, B, axes=0).reshape(-1, 3).tolist()


def get_dim(mat: Union[List[float], List[List[float]]]) -> List[int]:
    """Returns the dimensions of a 2D matrix. If flat, returns 0 as first dimension."""
    arr = np.array(mat)
    return [0, arr.size] if arr.ndim == 1 else list(arr.shape)


def add_arrays(arr1: List[float], arr2: List[float]) -> List[float]:
    """Adds the values of two arrays and returns them in a new array"""
    return np.add(arr1, arr2).tolist()


def determinant33(m: List[List[float]]) -> float:
    """Calculates the determinant of a 3x3 matrix"""
    return np.linalg.det(m)


def invert33(m: List[List[float]]) -> Union[List[List[float]], None]:
    """Inverts a 3x3 matrix"""
    try:
        return np.linalg.inv(m).tolist()
    except np.linalg.LinAlgError:
        print("invert33(m): det(m) == 0. m cannot be inverted!")
        return None


def multiply_m33v3(m: List[List[float]], v: List[float]) -> List[float]:
    """Multiplies a 3x3 matrix with a 3D vector"""
    return np.dot(m, v).tolist()


def transpose33(m: List[List[float]]) -> List[List[float]]:
    """Transposes a 3x3 matrix"""
    return np.array(m).T.tolist()


def solve33(m: List[List[float]], v: List[float]) -> List[float]:
    """Solves a 3x3 linear system"""
    return np.linalg.solve(m, v).tolist()



import logging
from collections import defaultdict
from typing import Union, List
import traceback
from pathlib import Path

from pydantic import BaseModel, ValidationError

from ataims.output_aims import OutputAims
from ataims.util import get_output_instance
from ataims.model import OutputData
from ataims.exceptions import FHIOutputParserError


logger = logging.getLogger(__name__)


def parse_outputfile(
    filename: Union[str, Path], 
    as_set: bool = False
) -> Union[OutputData, List[BaseModel]]:
    """Parse an FHI-aims output file and return structured data.

    Args:
        filename: Path to the FHI-aims output file. Can be either a string path
            or Path object. Both absolute and relative paths are supported.
            Example paths:
                - "calculation.out"
                - "./results/calculation.out"
                - "/home/user/aims/calculation.out"
                - "C:\\Users\\name\\aims\\calculation.out"
        as_set: If True, returns a list of individual data models.
            If False, returns a single combined OutputData object.
            Defaults to True.

    Returns:
        Union[OutputData, List[BaseModel]]: 
            If as_set is False, returns an OutputData object containing all parsed data.
            If as_set is True, returns a List of individual BaseModel objects.

    Raises:
        ValueError: If the file doesn't have a .out extension
        FileNotFoundError: If the specified file doesn't exist
        PermissionError: If the file can't be accessed due to permissions
        IOError: If there's an error reading the file
        FHIOutputParserError: If there's an error parsing the FHI-aims output

    Examples:
        >>> data = parse_outputfile("calculation.out")
        >>> # Using pathlib
        >>> from pathlib import Path
        >>> path = Path("aims/outputs/calculation.out")
        >>> data = parse_outputfile(path)
    """
    # Convert to Path object for OS-independent handling
    path = Path(filename)

    # Check extension
    if path.suffix != '.out':
        raise ValueError(f"File must have a .out extension, got: {path}")

    # Resolve to absolute path to handle relative paths
    path = path.resolve()

    try:
        with path.open('r', encoding='utf-8') as f:
            data = f.read()
    except (FileNotFoundError, PermissionError, IOError) as e:
        raise


    files_data = [{'name': str(filename), 'type': 'text/plain', 'content': data}]
    output = get_output_instance(files_data)
    output.parse_output_file(str(filename))

    try:
        pydantic_class = _output_to_pydantic_class(output)
    except Exception as e:
        error_message = f"Error in FHI-aims extraction: {str(e)}\n"
        if output.errors_set:
            error_message += (max(len(k) for k in output.errors_set) * '*') + '\n'
            error_message += '\n'.join(output.errors_set)
            error_message += '\n' + (max(len(k) for k in output.errors_set) * '*') + '\n'
        error_message += f"Traceback:\n{''.join(traceback.format_tb(e.__traceback__))}"
        raise FHIOutputParserError(error_message)

    if as_set:
        return [
            pydantic_class.results,
            pydantic_class.calculation_summary,
            pydantic_class.maximum_force_component,
            pydantic_class.change_of_sum_of_eigenvalues,
        ]
    else:
        return pydantic_class


def _output_to_pydantic_class(output: OutputAims) -> OutputData:
    """
    Converts the output of the outputparser (OutputAims instance) to a pydantic class.

    This function extracts various pieces of information from the OutputAims instance
    and organizes them into a structured format that can be validated using the
    OutputData pydantic model.

    Args:
        output (OutputAims): An instance of OutputAims containing parsed output data.

    Returns:
        OutputData: A validated pydantic model containing the structured output data.

    Raises:
        FHIOutputParserError: If there's an error during the extraction of FHI-aims data.
        ValidationError: If the extracted data fails to validate against the OutputData model.
        Exception: For any unexpected errors during the validation process.
    """
    data = defaultdict(dict)
    # System information
    data['system_information'] = output.system_information

    # Results 
    data['results'] = output.get_results_quantities()

    # Calculation summary
    output.get_calculation_info()
    data['calculation_summary']['code_version'] = output.calculation_info['codeVersion']['value']
    data['calculation_summary']['commit_number'] = output.calculation_info['commitNumber']['value']
    data['calculation_summary']['number_of_tasks'] = output.calculation_info['numberOfTasks']['value']
    # summary data can be missing due to the analysis not completing normally
    data['calculation_summary']['peak_memory_among_tasks_mb'] = output.memory['peakMemory']['value']
    data['calculation_summary']['largest_tracked_array_allocation_mb'] = output.memory.get('largestArray', {}).get('value')
    data['calculation_summary']['total_time'] = output.final_timings['totalTime']['value'] if output.final_timings else None

    # maxmimum force component
    data['maximum_force_component'] = output.relaxation_series if hasattr(output, 'relaxation_series') else None

    # eigenvalues, totalenergy, chargedensity
    output.get_data_series()
    data['change_of_sum_of_eigenvalues'] = output.data_series
    # errors
    terminated_normally_bool = output.exit_mode['normalTermination']
    data['calculation_summary']['calculation_exited_regularly'] = "Yes" if terminated_normally_bool else "No"
    if not terminated_normally_bool:
        output.scan_for_errors()  # added error parsing
        if output.errors_set:
            raise Exception("Critical errors detected in output file")


    # validation
    try:
        validated_data = OutputData.model_validate(data)
    except ValidationError as e:
        logger.error(f"Validation error: {e}")
        raise e
    except Exception as e:
        logger.error(f"Unexpected error validating output data: {e}")
        raise e

    return validated_data



from typing import List, Dict, Any, Optional, Generator, Match, Tuple
import re
import logging

from ataims.output import Output
from ataims.structure import Structure
from ataims import common_util as util


HARTREE = 27.211386245988
BOHR = 0.529177249

float_regex = r'[-+]?[0-9]+[.][0-9]*([eE][-+]?[0-9]+)?'

logger = logging.getLogger(__name__)


def parse_float_2_aa(float_string: str) -> float:
    return BOHR * float(float_string)


def parse_float_2_ev(float_string: str) -> float:
    return HARTREE * float(float_string)


class OutputExciting(Output):
    # warning: untested
    def __init__(self):
        super().__init__()
        self.inputs: Dict[str, List[str]] = {'InputXML': []}
        self.structure_in = Structure()
        self.structure_in.lat_vectors = []
        self.system_information: Dict[str, Any] = {'formula_unit': '', 'n_atoms': 0}
        self.output: Dict[str, List[str]] = {'preamble': [], 'initialization': [], 'groundstate': [], 'finale': []}
        self.bs_data: List[Dict[str, Any]] = []

    def parse_files(self, files_data: List[Dict[str, Any]]) -> None:
        for file_data in files_data:
            if file_data['type'] in ['application/gzip', 'application/x-gzip']:
                logger.info('Workflow output for Exciting is not yet implemented')
                return
            file_name = file_data['name']
            lines = util.get_tokenized_lines(file_data['content'])

            if file_name == 'dos.xml':
                self.files.dos[file_name] = get_dos_data_exciting(file_data['content'])
                self.filled_up = True
            elif file_name == 'bandstructure.xml':
                self.bs_data = get_bs_data_exciting(file_data['content'])
                self.filled_up = True
            elif file_name == 'INFO.OUT':
                self.files.output[file_name] = file_data['content']
                self.filled_up = True
            elif file_name == 'input.xml':
                self.files.input[file_name] = file_data['content']

    def get_bs_info(self) -> Dict[str, Any]:
        return {'bs_data': self.bs_data, 'lacking_segments': []}

    def get_relaxation_series(self) -> None:
        iters, data_i, force_i, d_end, ev_i = [], [], [], 0.0, []
        for i, loop in enumerate(self.scf_loops, 1):
            if loop['is_converged']:
                iters.append(i)
                d_end = loop['final_scf_energies']['total_energy']
                data_i.append(loop['final_scf_energies']['total_energy'])
                force_i.append(loop['max_force_component'])

        self.relaxation_series = {
            'labels': iters,
            'energy': {
                "label": "Total Energy",
                "data": [abs(value - d_end) for value in data_i],
                "yAxisID": 'yscenergy',
                "fill": False,
                "borderColor": "rgb(42,45,52)",
                "lineTension": 0.1
            },
            'forces': {
                "label": "Maximum Force Component",
                "data": [abs(value) for value in force_i],
                "yAxisID": 'ymaxforce',
                "fill": False,
                "borderColor": "rgb(0,157,220)",
                "lineTension": 0.1
            }
        }

    def get_data_series(self) -> None:
        data_series = []
        for loop in self.scf_loops:
            if loop.get('iterations'):
                data_series_iteration = {
                    'effective_potential': {'label': 'Change of effective Potential', 'color': 'rgb(0,157,220)', 'data': []},
                    'total_energy': {'label': 'Change of Total Energy', 'color': 'rgb(242,100,48)', 'data': []},
                    'charge_distance': {'label': 'Charge Distance', 'color': 'rgb(42,45,52)', 'data': []},
                }
                if self.run_time_choices.get('has_forces'):
                    data_series_iteration['forces'] = {
                        'label': 'Forces', 'color': 'rgb(42,45,104)', 'data': []
                    }

                for iteration in loop['iterations']:
                    for key, entry in data_series_iteration.items():
                        if iteration.get('convergence_accuracy'):
                            entry['data'].append(iteration['convergence_accuracy'][key])

                data_series.append(data_series_iteration)

        self.data_series = data_series

    def get_results_quantities(self) -> Dict[str, float]:
        from util import get_parallelepiped_volume
        quantities = {}
        last_converged = -1
        for i, loop in enumerate(self.scf_loops):
            if loop['is_converged']:
                last_converged = i

        if last_converged >= 0:
            last_loop = self.scf_loops[last_converged]
            quantities['Total Energy (eV)'] = last_loop['final_scf_energies']['total_energy']

            if self.run_time_choices['calculation_type'] != 'relaxation':
                last_iter = last_loop['iterations'][-1]
                for key, entry in last_iter['electron_info'].items():
                    quantities[entry['info']] = entry['value']

            if self.run_time_choices.get('is_periodic'):
                quantities['Cell Volume (Å³)'] = get_parallelepiped_volume(self.structure_in.lat_vectors)

            if self.run_time_choices['calculation_type'] == 'relaxation':
                from util import geo_text_from_structure
                final_structure_text = geo_text_from_structure(last_loop['structure'])
                href = util.create_object_url(final_structure_text, 'text/plain')
                quantities['download-link'] = href

        return quantities

    def get_calculation_info(self) -> Dict[str, Dict[str, Any]]:
        quantities = {}
        info_objects = [self.calculation_info, self.final_timings, self.exit_mode]
        for info_object in info_objects:
            if info_object:
                for key, entry in info_object.items():
                    quantities[entry['info']] = entry['value']
        return quantities

    def get_input_files_map(self) -> Dict[str, str]:
        return self.files.input

    def parse_output_file(self, file_name: Optional[str] = None) -> None:
        if not self.files.output:
            return

        if not file_name:
            file_name = next(iter(self.files.output))

        file_text = self.files.output[file_name]
        self.parse_file(file_name, file_text)

    def parse_file(self, file_name: str, file_text: str) -> None:
        self.normal_parser(file_name, file_text)
        self.get_system_information()
        self.get_data_series()
        if self.run_time_choices['calculation_type'] == 'relaxation':
            self.get_relaxation_series()

    def normal_parser(self, file_name: str, file_text: str) -> None:
        # Implementation of normal_parser goes here
        # This function is quite long, so I'll provide a skeleton and you can fill in the details

        def lines_iterator(body: List[str]) -> Generator[str, None, None]:
            for line in body:
                yield line

        def parse_until(regexs: List[re.Pattern], stop_re: re.Pattern) -> Optional[List[str]]:
            lines = []
            for line in line_it:
                for regex in regexs:
                    if regex.search(line):
                        lines.append(line)
                if stop_re.search(line):
                    return lines
            self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
            return None

        def wait_for(wait_re: re.Pattern) -> Optional[str]:
            for line in line_it:
                if wait_re.search(line):
                    return line
            self.errors.append([f"Did not find the following expression {wait_re.pattern}"])
            return None

        def check_for_until(check_re: re.Pattern, stop_re: re.Pattern) -> Optional[bool]:
            for line in line_it:
                if check_re.search(line):
                    return True
                elif stop_re.search(line):
                    return False
            self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
            return None

        def match_next_line(regex: re.Pattern) -> Optional[re.Match]:
            line = next(line_it, None)
            if line is not None:
                return regex.search(line)
            return None

        def get_lines_in_between(start_re: re.Pattern, stop_re: re.Pattern) -> Optional[List[str]]:
            lines = []
            line = wait_for(start_re)
            if line is None:
                return None
            for line in line_it:
                if stop_re.search(line):
                    return lines
                lines.append(line)
            self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
            return None

        def get_calculation_info() -> Optional[Dict[str, Dict[str, str]]]:
            line = wait_for(re.compile(r'\sEXCITING'))
            if line is None:
                return None
            version = line.split()[2]
            line = wait_for(re.compile(r'version hash id'))
            if line is None:
                return None
            hash_id = line.split()[4]
            return {
                'version': {'value': version, 'info': 'Version'},
                'hashID': {'value': hash_id[:9], 'info': 'Hash Id'}
            }

        def get_calculation_type() -> None:
            rtc = self.run_time_choices
            rtc['is_periodic'] = True
            lines = parse_until(
                [re.compile(r'Ground-state run'), re.compile(r'Structural optimisation')],
                re.compile(r'Starting initialization')
            )
            if lines is None:
                return
            if re.search(r'Ground-state run', lines[0]):
                rtc['calculation_type'] = 'singlePoint'
            elif re.search(r'Structural optimisation', lines[0]):
                rtc['calculation_type'] = 'relaxation'
            if rtc['calculation_type'] == 'relaxation':
                rtc['has_forces'] = True

        def parse_initialization() -> None:
            nonlocal current_structure
            line = wait_for(re.compile(r'\s+Lattice vectors'))
            if line is None:
                return
            init_structure = Structure()
            init_structure.lat_vectors = []
            for _ in range(3):
                match = match_next_line(re.compile(float_regex))
                if match:
                    init_structure.lat_vectors.append([parse_float_2_aa(x) for x in match.groups()])

            lines = get_lines_in_between(re.compile(r'Brillouin zone volume'), re.compile(r'Total number of atoms per unit cell'))
            if lines is None:
                return

            is_position_block = False
            is_fract = False
            species = ''

            for line in lines:
                if re.search(r'Species\s', line):
                    species = re.search(r'\((.*)\)', line).group(1)
                elif re.search(r'atomic positions', line) and not is_position_block:
                    is_fract = re.search(r'\(([^)]+)\)', line).group(1) == "lattice"
                    is_position_block = True
                elif is_position_block:
                    position = re.findall(float_regex, line)
                    if position:
                        if is_fract:
                            init_structure.add_atom_data([float(x) for x in position], species.strip(), is_fract)
                        else:
                            init_structure.add_atom_data([parse_float_2_aa(x) for x in position], species.strip(), is_fract)
                    else:
                        is_position_block = False

            self.structure_in = init_structure
            current_structure = init_structure

            line = wait_for('Spin treatment')
            if line:
                spin = line.split()[4]
                if spin == 'spin-unpolarised':
                    self.run_time_choices['spin'] = 'none'
                if spin == 'spin-polarised':
                    self.run_time_choices['spin'] = 'collinear'

            line = wait_for('Ending initialization')
            if line is None:
                return

        def parse_until(regexs: List[re.Pattern], stop_re: re.Pattern) -> Optional[List[str]]:
            lines = []
            for line in line_it:
                for regex in regexs:
                    if regex.search(line):
                        lines.append(line)
                if stop_re.search(line):
                    return lines
            self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
            return None

        def wait_for(wait_re: re.Pattern) -> Optional[str]:
            for line in line_it:
                if wait_re.search(line):
                    return line
            self.errors.append([f"Did not find the following expression {wait_re.pattern}"])
            return None

        def check_for_until(check_re: re.Pattern, stop_re: re.Pattern) -> Optional[bool]:
            for line in line_it:
                if check_re.search(line):
                    return True
                elif stop_re.search(line):
                    return False
            self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
            return None

        def match_next_line(regex: re.Pattern) -> Optional[Match[str]]:
            line = next(line_it, None)
            if line is not None:
                return regex.search(line)
            return None

        def get_lines_in_between(start_re: re.Pattern, stop_re: re.Pattern) -> Optional[List[str]]:
            lines = []
            line = wait_for(start_re)
            if line is None:
                return None
            for line in line_it:
                if stop_re.search(line):
                    return lines
                lines.append(line)
            self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
            return None

        def get_calculation_info() -> Optional[Dict[str, Dict[str, Any]]]:
            line = wait_for(re.compile(r'\sEXCITING'))
            if line is None:
                return None
            version = line.split()[2]
            line = wait_for(re.compile(r'version hash id'))
            if line is None:
                return None
            hash_id = line.split()[4]
            return {
                'version': {'value': version, 'info': 'Version'},
                'hashID': {'value': hash_id[:9], 'info': 'Hash Id'}
            }

        def get_calculation_type() -> None:
            rtc = self.run_time_choices
            rtc['is_periodic'] = True
            lines = parse_until(
                [re.compile(r'Ground-state run'), re.compile(r'Structural optimisation')],
                re.compile(r'Starting initialization')
            )
            if lines is None:
                return
            if re.search(r'Ground-state run', lines[0]):
                rtc['calculation_type'] = 'singlePoint'
            elif re.search(r'Structural optimisation', lines[0]):
                rtc['calculation_type'] = 'relaxation'
            if rtc['calculation_type'] == 'relaxation':
                rtc['has_forces'] = True

        def parse_initialization() -> None:
            global current_structure
            line = wait_for(re.compile(r'\s+Lattice vectors'))
            if line is None:
                return
            init_structure = Structure()
            init_structure.lat_vectors = []
            for _ in range(3):
                match = match_next_line(re.compile(float_regex))
                if match:
                    init_structure.lat_vectors.append([parse_float_2_aa(x) for x in match.groups()])

            lines = get_lines_in_between(re.compile(r'Brillouin zone volume'), re.compile(r'Total number of atoms per unit cell'))
            if lines is None:
                return

            is_position_block = False
            is_fract = False
            species = ''

            for line in lines:
                if re.search(r'Species\s', line):
                    species = re.search(r'\((.*)\)', line).group(1)
                elif re.search(r'atomic positions', line) and not is_position_block:
                    is_fract = re.search(r'\(([^)]+)\)', line).group(1) == "lattice"
                    is_position_block = True
                elif is_position_block:
                    position = re.findall(float_regex, line)
                    if position:
                        if is_fract:
                            init_structure.add_atom_data([float(x) for x in position], species.strip(), is_fract)
                        else:
                            init_structure.add_atom_data([parse_float_2_aa(x) for x in position], species.strip(), is_fract)
                    else:
                        is_position_block = False

            self.structure_in = init_structure
            current_structure = init_structure

            line = wait_for('Spin treatment')
            if line:
                spin = line.split()[4]
                if spin == 'spin-unpolarised':
                    self.run_time_choices['spin'] = 'none'
                if spin == 'spin-polarised':
                    self.run_time_choices['spin'] = 'collinear'

            line = wait_for('Ending initialization')
            if line is None:
                return

        def get_scf_loops(run_time_choices: Dict[str, Any]) -> Optional[List[Dict[str, Any]]]:
            line = wait_for('Self-consistent loop started')
            if line is None:
                return None
            is_converged, iterations = get_iterations(run_time_choices)
            first_scf_loop = {
                'is_converged': is_converged,
                'iterations': iterations,
                'structure': current_structure,
                'final_scf_energies': {}
            }
            if not is_converged:
                return [first_scf_loop]
            last_iteration = get_last_iteration()
            if not last_iteration:
                return [first_scf_loop]
            first_scf_loop['iterations'].append(last_iteration)
            first_scf_loop['final_scf_energies']['total_energy'] = last_iteration['scf_energies']['total_energy']
            if run_time_choices['calculation_type'] == 'relaxation':
                line = wait_for(re.compile(r'Maximum force magnitude'))
                if line is None:
                    return [first_scf_loop]
                first_scf_loop['max_force_component'] = float(re.search(float_regex, line).group()) * HARTREE / BOHR
            scf_loops = [first_scf_loop]
            if run_time_choices['calculation_type'] == 'relaxation':
                is_converged, relaxation_scf_loops = get_relaxation_scf_loops()
                scf_loops = [first_scf_loop] + relaxation_scf_loops
            return scf_loops
        
        def get_relaxation_scf_loops() -> Tuple[bool, List[Dict[str, Any]]]:
            """
            Get the relaxation SCF loops from the output.

            Returns:
                Tuple[bool, List[Dict[str, Any]]]: A tuple containing:
                    - bool: True if the relaxation converged, False otherwise
                    - List[Dict[str, Any]]: List of optimization steps
            """
            relaxation_scf_loops: List[Dict[str, Any]] = []
            is_not_converged = True
            while is_not_converged:
                is_not_converged = check_for_until(r'Optimization step', r'Force convergence target achieved')
                if is_not_converged is False:
                    return True, relaxation_scf_loops
                elif is_not_converged is None:
                    return False, relaxation_scf_loops
                else:
                    optimization_step = get_optimization_step()
                    if not optimization_step:
                        return False, relaxation_scf_loops
                    relaxation_scf_loops.append(optimization_step)
            
            # This line should never be reached, but we include it for completeness
            return False, relaxation_scf_loops

        def get_optimization_step() -> Optional[Dict[str, Any]]:
            """
            Get the optimization step information from the output.

            Returns:
                Optional[Dict[str, Any]]: A dictionary containing optimization step information,
                or None if the required information is not found.
            """
            optimization_step: Dict[str, Any] = {
                'final_scf_energies': {},
                'max_force_component': None,
                'is_converged': True
            }

            line = wait_for(re.compile(r'Maximum force magnitude'))
            if not line:
                return None
            optimization_step['max_force_component'] = float(re.search(float_regex, line).group()) * HARTREE / BOHR

            line = wait_for(re.compile(r'Total energy at this optimization step'))
            if not line:
                return None
            optimization_step['final_scf_energies']['total_energy'] = float(re.search(float_regex, line).group())

            lines = get_lines_in_between(re.compile(r'Atomic positions at this step'), re.compile(r'Total atomic forces'))
            if not lines:
                return None

            updated_structure = Structure()
            updated_structure.lat_vectors = current_structure.lat_vectors
            for line in lines:
                if 'atom' in line:
                    position = re.findall(float_regex, line)
                    species = line.split()[2]
                    updated_structure.add_atom_data([float(pos) for pos in position], species.strip(), True)

            optimization_step['structure'] = updated_structure
            return optimization_step

        def get_iterations(run_time_choices: Dict[str, Any]) -> Tuple[bool, List[Dict[str, Any]]]:
            """
            Get iterations from the output.

            Args:
                run_time_choices (Dict[str, Any]): Runtime choices for the calculation.

            Returns:
                Tuple[bool, List[Dict[str, Any]]]: A tuple containing a boolean indicating convergence
                and a list of iteration data.
            """
            iterations: List[Dict[str, Any]] = []
            is_not_converged = True
            while is_not_converged:
                is_not_converged = check_for_until(re.compile(r'SCF iteration number'), re.compile(r'Convergence targets achieved\.'))
                if is_not_converged is False:
                    return True, iterations
                elif is_not_converged is None:
                    return False, iterations
                else:
                    iteration = get_iteration(run_time_choices)
                    if not iteration:
                        return False, iterations
                    iterations.append(iteration)
            return False, iterations  # This line should never be reached

        def get_iteration(run_time_choices: Dict[str, Any]) -> Optional[Dict[str, Any]]:
            """
            Get data for a single iteration.

            Args:
                run_time_choices (Dict[str, Any]): Runtime choices for the calculation.

            Returns:
                Optional[Dict[str, Any]]: A dictionary containing iteration data, or None if required information is not found.
            """
            iteration: Dict[str, Any] = {
                'scf_energies': {},
                'electron_info': {}
            }

            line = wait_for(re.compile(r'Total energy'))
            if not line:
                return None
            iteration['scf_energies']['total_energy'] = parse_float_2_ev(re.search(float_regex, line).group())

            line = wait_for(re.compile(r'Fermi energy'))
            if not line:
                return None
            iteration['electron_info']['fermi_energy'] = {
                'value': parse_float_2_ev(re.search(float_regex, line).group()),
                'info': 'Fermi Energy (eV)'
            }

            conv_acc = get_conv_acc(run_time_choices)
            if not conv_acc:
                return None
            iteration['convergence_accuracy'] = conv_acc

            return iteration

        def get_last_iteration() -> Optional[Dict[str, Any]]:
            """
            Get data for the last iteration.

            Returns:
                Optional[Dict[str, Any]]: A dictionary containing the last iteration data,
                or None if required information is not found.
            """
            iteration: Dict[str, Any] = {
                'scf_energies': {},
                'electron_info': {}
            }

            line = wait_for(re.compile(r'Total energy'))
            if not line:
                return None
            iteration['scf_energies']['total_energy'] = parse_float_2_ev(re.search(float_regex, line).group())

            line = wait_for(re.compile(r'Fermi energy'))
            if not line:
                return None
            iteration['electron_info']['fermi_energy'] = {
                'value': parse_float_2_ev(re.search(float_regex, line).group()),
                'info': 'Fermi Energy (eV)'
            }

            line = wait_for(re.compile(r'Estimated fundamental gap'))
            if not line:
                return None
            iteration['electron_info']['gap'] = {
                'value': parse_float_2_ev(re.search(float_regex, line).group()),
                'info': 'Estimated fundamental gap (eV)'
            }

            return iteration

        def get_conv_acc(run_time_choices: Dict[str, Any]) -> Optional[Dict[str, float]]:
            """
            Get convergence accuracy data.

            Args:
                run_time_choices (Dict[str, Any]): Runtime choices for the calculation.

            Returns:
                Optional[Dict[str, float]]: A dictionary containing convergence accuracy data,
                or None if required information is not found.
            """
            conv_acc: Dict[str, float] = {}

            line = wait_for(re.compile(r'RMS change in effective potential'))
            if not line:
                return None
            conv_acc['effective_potential'] = float(re.search(float_regex, line).group())

            line = wait_for(re.compile(r'Absolute change in total energy'))
            if not line:
                return None
            conv_acc['total_energy'] = float(re.search(float_regex, line).group())

            line = wait_for(re.compile(r'Charge distance'))
            if not line:
                return None
            conv_acc['charge_distance'] = float(re.search(float_regex, line).group())

            if run_time_choices.get('has_forces'):
                line = wait_for(re.compile(r'Abs. change in max-nonIBS-force'))
                if not line:
                    return None
                conv_acc['forces'] = float(re.search(float_regex, line).group())

            return conv_acc

        def get_final_timings() -> Optional[Dict[str, Dict[str, Any]]]:
            line = wait_for(re.compile(r'Total time spent'))
            if line is None:
                return None
            return {
                'total_time': {
                    'value': float(re.search(float_regex, line).group()),
                    'info': 'Total Time (s)'
                }
            }

        def get_exit_mode() -> Dict[str, Dict[str, str]]:
            line = wait_for(re.compile(r'EXCITING.*stopped'))
            value = 'yes' if line else 'no'
            return {
                'exit_mode': {
                    'value': value,
                    'info': 'Calculation exited regularly'
                }
            }

        current_structure = None
        lines = file_text.split('\n')
        line_it = lines_iterator(lines)

        self.calculation_info = get_calculation_info()
        get_calculation_type()
        parse_initialization()
        self.scf_loops = get_scf_loops(self.run_time_choices)

        self.final_timings = get_final_timings()
        self.exit_mode = get_exit_mode()

    def get_system_information(self) -> None:
        n_atoms = len(self.structure_in.atoms)
        species = {}
        for atom in self.structure_in.atoms:
            sp = atom.species
            species[sp] = species.get(sp, 0) + 1
        
        formula_unit = ''.join(f"{key}{num}" if num > 1 else key for key, num in species.items())
        
        self.system_information['n_atoms'] = n_atoms
        self.system_information['formula_unit'] = formula_unit

def get_dos_data_exciting(xml_text: str) -> List[List[float]]:
    # Implementation goes here
    pass

def get_bs_data_exciting(xml_text: str) -> List[Dict[str, Any]]:
    # Implementation goes here
    pass

def get_kpoint_coords(vertexes: List[Dict[str, Any]], index: int, dist_per_unit: float) -> List[float]:
    # Implementation goes here
    pass



from typing import List, Dict, Any, Optional, Tuple, Generator, Union
import re
import logging

from ataims.output import Output
from ataims.structure import Structure
from ataims import common_util as util
# import output_parser.user_msg_box as UserMsgBox


logger = logging.getLogger(__name__)


class OutputAims(Output):
    def __init__(self):
        super().__init__()
        __slots__ = [
            'inputs',
            'output',
            'no_soc',
            'is_gw',
            'segments_map',
            'band_names',
            'float_regex',
            'errors_set'
        ]

        self.inputs: Dict[str, List[str]] = {'controlIn': [], 'geometryIn': []}
        self.output: Dict[str, List[str]] = {'body': [], 'finale': []}
        self.no_soc: bool = False
        self.is_gw: bool = False
        self.segments_map: Dict[str, Any] = {}
        self.band_names: List[str] = []
        self.float_regex = r'-?\d+\.\d+(?:[eE][-+]?\d+)?'
        # error lines are redundant
        self.errors_set = set()

    def parse_files(self, files_data: List[Dict[str, Any]]) -> 'OutputAims':
        if len(files_data) == 1 and 'gzip' in files_data[0]['type']:
            # TODO: replace this method
            pass
            # data = await self.prepare_workflow_outputs(files_data[0])
            # return self._parse(data, True)
        else:
            return self._parse(files_data, False)

    async def prepare_workflow_outputs(self, file_data: Dict[str, Any]) -> Dict[str, Any]:
        response = await util.fetch('/upload-workflow', method='POST', body=file_data['content'])
        if not response.ok:
            # UserMsgBox.set_error('Unknown server ERROR')
            return {}
        text = await response.text()
        return util.json_loads(text)

    def _parse(self, files_data: List[Dict[str, Any]], is_workflow: bool = False) -> 'OutputAims':
        import os
        for file_data in files_data:
            file_name = file_data['name']
            is_band_file = 'band' in file_name if is_workflow else file_name.startswith('band')
            is_gw_band_file = 'GW_band' in file_name if is_workflow else file_name.startswith('GW_band')
            if is_band_file and file_name.endswith('.out'):
                self.band_names.append(file_name)
            if is_band_file and file_name.endswith('.no_soc'):
                self.band_names.append(file_name)
                self.no_soc = True
            if is_gw_band_file and file_name.endswith('.out'):
                self.is_gw = True
                self.band_names.append(file_name)

        logger.info(f"{self.band_names}: no SOC - {self.no_soc}, is GW - {self.is_gw}")

        for file_data in files_data:
            file_name = file_data['name']
            workflow_idx = None
            if is_workflow:
                workflow = os.path.dirname(file_name)
                if workflow not in self.workflows:
                    self.workflows.append(workflow)
                workflow_idx = self.workflows.index(workflow)
            is_band_file = 'band' in file_name if is_workflow else file_name.startswith('band')
            is_gw_band_file = 'GW_band' in file_name if is_workflow else file_name.startswith('GW_band')
            lines = util.get_tokenized_lines(file_data['content'])

            if re.match(r'KS_DOS_total(_tetrahedron)?.dat', file_name) or \
               re.match(r'_proj_dos(_tetrahedron)?\.dat(\.no_soc)?', file_name) or \
               re.match(r'atom_proj(ected)?_dos(_tetrahedron)?_[A-Z][a-z]?[0-9]{4}\.dat(\.no_soc)?', file_name):
                self.files['dos'][file_name] = self.get_tabulated_data_fhi_aims(lines)
                self.filled_up = True
            elif re.match(r'_proj_dos(_tetrahedron)?_spin_(up|dn)\.dat(\.no_soc)?', file_name):
                alt_file_name = file_name.replace('_spin_up', '').replace('_spin_dn', '')
                if alt_file_name in self.files['dos']:
                    data = self.files['dos'][alt_file_name]
                    data_new = self.get_tabulated_data_fhi_aims(lines)
                    for i, line in enumerate(data_new):
                        if 'spin_up' in file_name:
                            data[i][1] = line[1]
                        else:
                            data[i][2] = line[1]
                else:
                    data = []
                    for line in self.get_tabulated_data_fhi_aims(lines):
                        if 'spin_up' in file_name:
                            data.append([line[0], line[1], 0.0])
                        else:
                            data.append([line[0], 0.0, line[1]])
                    self.files['dos'][alt_file_name] = data
                self.filled_up = True
                self.run_time_choices['has_spin'] = True
            elif re.match(r'absorption(_soc)?(_Lorentzian)?(_Tetrahedron)?_[0-9].[0-9]{4}(_x_x)?(_y_y)?(_z_z)?.out', file_name):
                self.files['absorption'][file_name] = self.get_tabulated_data_fhi_aims(lines)
                self.filled_up = True
            elif re.match(r'dielectric_function(_soc)?(_Lorentzian)?(_Tetrahedron)?_[0-9].[0-9]{4}(_x_x)?(_y_y)?(_z_z)?(_x_y)?(_x_z)?(_y_z)?.out', file_name):
                self.files['dielectric'][file_name] = self.get_tabulated_data_fhi_aims(lines)
                self.filled_up = True
            elif (is_band_file or is_gw_band_file) and file_name.endswith('.out'):
                self.add_bs_file_data(file_name, self.segments_map, lines, is_gw_band_file, workflow_idx)
                self.filled_up = True
            elif 'control.in' in file_name:
                self.files['input'][file_name] = file_data['content']
                self.control_in = lines
            elif 'geometry.in' in file_name:
                self.files['input'][file_name] = file_data['content']
            elif 'Invoking FHI-aims ...' in file_data['content']:
                self.files['output'][file_name] = file_data['content']
                self.filled_up = True

        return self

    @staticmethod
    def get_tabulated_data_fhi_aims(lines: List[str]) -> List[List[float]]:
        data = []
        for line in lines:
            if not line.startswith('#') and len(line.strip()) > 1:
                a = [float(x) for x in line.strip().split()]
                data.append(a)
        return data

    @staticmethod
    def add_bs_file_data(full_file_name: str, segments_map: Dict[str, Any], lines: List[str], is_gw: bool, index: Optional[int] = None) -> None:
        import os
        file_name = os.path.basename(full_file_name)
        segment_num_ind = [5, 8] if not is_gw else [8, 11]
        spin_index_ind = 4 if not is_gw else 7
        segment_num = file_name[segment_num_ind[0]:segment_num_ind[1]]
        spin_index = 0
        if index is None:
            if file_name.endswith(".no_soc"):
                spin_index = int(file_name[spin_index_ind])
            elif file_name.endswith(".out"):
                spin_index = int(file_name[spin_index_ind]) - 1
        else:
            spin_index = index

        if segment_num not in segments_map:
            segments_map[segment_num] = {
                'segment_num': int(segment_num),
                'band_k_points': [],
                'band_energies': [[], [], []],
                'bs_name': [[], [], []],
                'segment_name': [[], [], []]
            }
        segment_data = segments_map[segment_num]

        file_name_token = file_name[:segment_num_ind[0]] + '***' + file_name[segment_num_ind[1]:]
        segment_data['bs_name'][spin_index].append(file_name_token)
        segment_data['segment_name'][spin_index].append(file_name)

        k_point_flag = spin_index if len(segment_data['band_k_points']) == 0 else None
        for line in lines:
            if len(line.strip()) > 1:
                a = [float(x) for x in line.strip().split()]
                if spin_index == k_point_flag:
                    segment_data['band_k_points'].append(a[1:4])
                segment_data['band_energies'][spin_index].append([e for i, e in enumerate(a[4:]) if i % 2 == 1])

    def get_bs_info(self) -> Dict[str, Any]:
        if not hasattr(self, 'bs_info'):
            if self.control_in:
                self.bs_info = self.find_lacking_segments(self.control_in, self.segments_map)
            else:
                self.bs_info = self.find_lacking_segments(None, self.segments_map)
        self.bs_info['has_spin'] = (self.run_time_choices['spin'] == "collinear")
        self.bs_info['has_soc'] = self.run_time_choices['has_soc']
        return self.bs_info

    def parse_output_file(self, file_name: Optional[str] = None) -> None:
        if len(self.files['output']) == 0:
            return

        control_in_lines = self.control_in

        if not file_name:
            file_name = next(iter(self.files['output']))

        file_text = self.files['output'][file_name]
        self.parse_file(file_name, file_text)

        if not control_in_lines:
            control_in_lines = self.inputs['controlIn']   
        # this finds and passes "'  # FHI-aims relaxation template'", unclear if it matters for our purposes
        self.bs_info = self.find_lacking_segments(control_in_lines, self.segments_map)

    @staticmethod
    def find_lacking_segments(control_in: Optional[List[str]], segments_map: Dict[str, Any]) -> Dict[str, Any]:
        lacking_segments = []
        bs_data = []

        if control_in:
            lacking_segments = OutputAims.add_segment_labels_to_bs(control_in, segments_map)

        segment_nums = sorted(segments_map.keys())
        for segment_num in segment_nums:
            bs_data.append(segments_map[segment_num])

        # Find missing segments
        if segment_nums:
            max_num = int(segment_nums[-1])
            for i in range(1, max_num + 1):
                code = OutputAims.get_file_code_from_number(i)
                if code not in segment_nums:
                    lacking_segments.append(code)

        return {'lacking_segments': lacking_segments, 'bs_data': bs_data}

    @staticmethod
    def add_segment_labels_to_bs(lines: List[str], segments_map: Dict[str, Any]) -> List[str]:
        segments_labels = []
        lacking_segments = []

        for line in lines:
            a = line.strip().split()
            if len(a) >= 4 and a[0] == 'output' and a[1] == 'band':
                segments_labels.append(a[-2:])

        for i, label in enumerate(segments_labels):
            key = OutputAims.get_file_code_from_number(i + 1)
            if key in segments_map:
                segments_map[key]['band_segm_labels'] = label
            else:
                lacking_segments.append(key)

        return lacking_segments

    @staticmethod
    def get_file_code_from_number(n: int) -> str:
        return f'00{n}' if n < 10 else f'0{n}'

    def parse_file(self, file_name: str, file_text: str) -> None:
        self.normal_parser(file_text)
        self.ks_ev = self.get_ks_ev(file_text)
        self.get_data_series()
        if self.run_time_choices['calculationType'] == 'relaxation':
            self.get_relaxation_series()

    def get_data_series(self) -> None:
        data_series = []

        for loop in self.scf_loops:
            has_iter = loop.get('iterations') and len(loop['iterations']) > 0
            if has_iter:
                data_series_iteration = {
                    'eigenvalues': {'label': 'Change of Eigenvalues', 'color': 'rgb(0,157,220)', 'data': []},
                    'totalEnergy': {'label': 'Change of Total Energy', 'color': 'rgb(242,100,48)', 'data': []},
                }
                if self.run_time_choices['spin'] == 'collinear':
                    data_series_iteration['chargeDensityUp'] = {'label': 'Change of Charge Density', 'color': 'rgb(42,45,52)', 'data': []}
                    data_series_iteration['chargeDensityDown'] = {'label': 'Change of Spin Density', 'color': 'rgb(84,45,52)', 'data': []}
                else:
                    data_series_iteration['chargeDensity'] = {'label': 'Change of Charge Density', 'color': 'rgb(42,45,52)', 'data': []}

                for iteration in loop['iterations']:
                    for key, entry in data_series_iteration.items():
                        entry['data'].append(iteration['convergenceAccuracy'][key])

                data_series.append(data_series_iteration)

        self.data_series = data_series

    def get_relaxation_series(self) -> None:
        iters, data_i, force_i = [], [], []
        d_end = 0.0
        logging.debug(f"Number of SCF loops: {len(self.scf_loops)}")
        for i, loop in enumerate(self.scf_loops, 1):
            logging.debug(f"Processing loop {i}")
            logging.debug(f"Loop keys: {loop.keys()}")
            logging.debug(f"isConverged: {loop.get('isConverged', 'Not found')}")
            if loop.get('isConverged', False):
                iters.append(i)
                if 'finalScfEnergies' in loop and 'totalEnergy' in loop['finalScfEnergies']:
                    energy = loop['finalScfEnergies']['totalEnergy']
                    data_i.append(energy)
                    if i == len(self.scf_loops):  # Last iteration
                        d_end = energy
                else:
                    logging.warning(f"Missing 'finalScfEnergies' or 'totalEnergy' in loop {i}")
                if 'maxForceComponent' in loop:
                    force_i.append(loop['maxForceComponent'])
                else:
                    logging.warning(f"Missing 'maxForceComponent' in loop {i}")
        logging.debug(f"Processed {len(iters)} converged iterations")

        if not iters:
            logging.warning("No converged iterations found. Check if this is a relaxation calculation.")
            return

        self.relaxation_series = {
            'labels': iters,
            'energy': {
                "label": "Total Energy",
                "data": [abs(value - d_end) for value in data_i],
                "yAxisID": 'yscenergy',
                "fill": False,
                "borderColor": "rgb(42,45,52)",
                "lineTension": 0.1
            },
            'forces': {
                "label": "Maximum Force Component",
                "data": [abs(force) for force in force_i],
                "yAxisID": 'ymaxforce',
                "fill": False,
                "borderColor": "rgb(0,157,220)",
                "lineTension": 0.1
            }
        }

    def get_structure(self) -> Optional[Structure]:
        from .util import line_array_to_text
        geometry_in_text = None
        if self.inputs['geometryIn']:
            geometry_in_text = line_array_to_text(self.inputs['geometryIn'])
        elif self.files['input'].get('geometry.in'):
            geometry_in_text = self.files['input']['geometry.in']

        if geometry_in_text:
            return util.get_structure_from_file_content('geometry.in', geometry_in_text)
        return None

    def get_results_quantities(self) -> Dict[str, Any]:
        quantities = {}
        last_converged = -1
        for loop in self.scf_loops:
            if loop['isConverged']:
                last_converged += 1

        if last_converged >= 0:
            last_loop = self.scf_loops[last_converged]
            last_iter = last_loop['iterations'][-1]
            quantities['Total Energy (eV)'] = last_loop['finalScfEnergies']['totalEnergy']

            for entry in last_iter['electronInfo'].values():
                quantities[entry['info']] = entry['value']

            if self.run_time_choices['is_periodic']:
                from .util import get_parallelepiped_volume
                quantities['Cell Volume (&#197;<sup>3</sup>)'] = float(get_parallelepiped_volume(self.structure_in.lat_vectors))

            if self.run_time_choices['calculationType'] == 'relaxation':
                from .util import geo_text_from_structure
                final_structure_text = geo_text_from_structure(last_loop['structure'])
                # not needed for now
                # href = util.create_blob_url(final_structure_text, 'text/plain')
                # quantities['download-link'] = href

        return quantities

    def get_input_files_map(self) -> Dict[str, str]:
        from .util import line_array_to_text
        input_files_map = {}
        if self.inputs['geometryIn']:
            input_files_map['geometry.in'] = line_array_to_text(self.inputs['geometryIn'])
        if self.inputs['controlIn']:
            input_files_map['control.in'] = line_array_to_text(self.inputs['controlIn'])

        for name, content in self.files['input'].items():
            if name not in input_files_map:
                input_files_map[name] = content

        return input_files_map

    def get_run_time_choices(self) -> None:
        rtc = self.run_time_choices
        rtc['output_level'] = 'normal'  # Set a default value
        for line in self.inputs['controlIn']:
            trim_line = line.strip()

            if trim_line.startswith('relax_geometry'):
                rtc['calculationType'] = 'relaxation'
                rtc['has_forces'] = True
            if trim_line.startswith('spin') and 'collinear' in trim_line:
                rtc['spin'] = 'collinear'
            if trim_line.startswith('output_level') and 'MD_light' in trim_line:
                rtc['output_level'] = 'MD_light'
            if trim_line.startswith('k_grid'):
                rtc['is_periodic'] = True
            if trim_line.startswith('relax_unit_cell'):
                rtc['has_stress'] = True
            if trim_line.startswith('include_spin_orbit'):
                rtc['has_soc'] = True
            if trim_line.startswith('output') and 'hirshfeld' in trim_line:
                rtc['has_hirshfeld'] = True
            if trim_line.startswith('many_body_dispersion'):
                rtc['has_mbd'] = True
            if trim_line.startswith('output') and 'mulliken' in trim_line:
                rtc['has_mulliken'] = True

    def get_ks_ev(self, file_text: str) -> List[List[float]]:
        ks_ev = []
        lines = file_text.split('\n')
        found_ks_ev = False
        spin_collinear = False
        for line in lines:
            if 'spin' in line and 'collinear' in line:
                spin_collinear = True
            if "State" in line and "Occupation" in line and "Eigenvalue" in line and not spin_collinear:
                ks_ev = []
                found_ks_ev = True
                continue  # Skip the header line
            elif spin_collinear and 'Spin-up' in line and 'eigenvalues:' in line:
                ks_ev = []
            if "State" in line and "Occupation" in line and "Eigenvalue" in line and spin_collinear:
                found_ks_ev = True
                continue  # Skip the header line
            elif found_ks_ev and line.strip():
                a = [float(x) for x in line.strip().split() if x]
                if len(a) >= 4:
                    ks_ev.append([a[0], a[1], a[3]])
                elif len(a) >= 3:
                    ks_ev.append([a[0], 0.000, a[2]])
            elif found_ks_ev and not line.strip():
                found_ks_ev = False
        return ks_ev

    def parse_until(self, reg_exs: List[re.Pattern], stop_re: re.Pattern) -> Optional[List[str]]:
        lines = []
        for line in self.line_it:
            if any(re.match(regex, line) for regex in reg_exs):
                lines.append(line)
            if re.match(stop_re, line):
                return lines
        self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
        return None

    def wait_for(self, wait_re: Union[str, re.Pattern]) -> Optional[str]:
        if isinstance(wait_re, str):
            wait_re = re.compile(wait_re, re.IGNORECASE)
        for line in self.line_it:
            if wait_re.search(line):
                return line
        self.errors.append([f"Did not find the following expression {wait_re.pattern}"])
        return None

    def match_next_line(self, regex: re.Pattern) -> Optional[re.Match]:
        line = next(self.line_it, None)
        return re.match(regex, line) if line else None

    def get_lines_in_between(self, start_re: re.Pattern, stop_re: re.Pattern) -> Optional[List[str]]:
        lines = []
        self.wait_for(start_re)
        for line in self.line_it:
            if re.match(stop_re, line):
                return lines
            lines.append(line)
        self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
        return None

    def parse_scf_energies(self) -> Optional[Dict[str, float]]:
        energy_regex = [
            (re.compile(r'Sum of eigenvalues'), 'sumEigenvalues'),
            (re.compile(r'XC energy correct'), 'XCenergyCorrection'),
            (re.compile(r'XC potential correct'), 'XCpotentialCorrection'),
            (re.compile(r'Free-atom electrostatic energy'), 'freeAtomElStat'),
            (re.compile(r'Hartree energy correct'), 'hartreeEnergyCorrection'),
            (re.compile(r'Entropy correct'), 'entropyCorrection'),
            (re.compile(r'Total energy\s'), 'totalEnergy'),
            (re.compile(r'Total energy, T\s'), 'totalEnergyCorrected'),
        ]
        scf_energies = {}
        line = self.wait_for(re.compile(r'Total energy components:'))
        if line is None:
            return None

        for regex, key in energy_regex:
            line = self.wait_for(regex)
            if line:
                m = re.findall(self.float_regex, line)
                if m and len(m) == 2:
                    scf_energies[key] = float(m[1])

        return {'scfEnergies': scf_energies} if len(scf_energies) == 8 else None

    def parse_conv_acc(self) -> Optional[Dict[str, float]]:
        conv_acc = {}
        line = self.wait_for(re.compile(r'Self-consistency convergence accuracy:'))
        if line is None:
            return None
        line = self.wait_for(re.compile(r'\| Change of charge'))
        if self.run_time_choices['spin'] == 'collinear':
            m = re.findall(self.float_regex, line)
            conv_acc['chargeDensityUp'] = float(m[0])
            conv_acc['chargeDensityDown'] = float(m[1])
        else:
            conv_acc['chargeDensity'] = float(re.findall(self.float_regex, line)[0])
        line = self.wait_for(re.compile(r'\| Change of sum of eigenvalues'))
        conv_acc['eigenvalues'] = float(re.findall(self.float_regex, line)[0])
        line = self.wait_for(re.compile(r'\| Change of total energy'))
        conv_acc['totalEnergy'] = float(re.findall(self.float_regex, line)[0])
        return {'convergenceAccuracy': conv_acc}

    def get_electron_info(self) -> Optional[Dict[str, Dict[str, Any]]]:
        electron_info = {}
        if self.run_time_choices['output_level'] == 'normal':
            line = self.wait_for(re.compile(r'\| Chemical potential'))
            if line is None:
                return None
            electron_info['fermiEnergy'] = {
                'value': float(re.findall(self.float_regex, line)[0]),
                'info': 'Fermi Energy (eV)'
            }

        line = self.wait_for(re.compile(r'Highest occupied state|Reaching maximum number'))
        if line is None:
            return None
        if 'Reached maximum number' in line:
            return {}
        state = float(re.findall(self.float_regex, line)[0])
        line = self.wait_for(re.compile(r'\| Occupation number:'))
        if line is None:
            return None
        occ_number = float(re.findall(self.float_regex, line)[0])
        electron_info['highestOccState'] = {
            'value': state,
            'info': 'Highest occupied state (eV)',
            'occNumber': occ_number
        }

        line = self.wait_for(re.compile(r'Lowest unoccupied state'))
        if line is None:
            return None
        state = float(re.findall(self.float_regex, line)[0])
        line = self.wait_for(re.compile(r'\| Occupation number:'))
        occ_number = float(re.findall(self.float_regex, line)[0])
        electron_info['lowestUnOccState'] = {
            'value': state,
            'info': 'Lowest unoccupied state (eV)',
            'occNumber': occ_number
        }

        line = self.wait_for(re.compile(r'verall HOMO-LUMO gap:'))
        if line is None:
            return None
        gap = float(re.findall(self.float_regex, line)[0])
        electron_info['gap'] = {
            'value': gap,
            'info': 'Estimated HOMO-LUMO gap (eV)',
        }
        return {'electronInfo': electron_info}

    def get_iteration(self) -> Tuple[Optional[bool], Optional[Dict[str, Any]]]:
        iteration = {}
        is_converged = False
        line = self.wait_for(r'(Begin self-consistency iteration|SELF-CONSISTENCY CYCLE DID NOT CONVERGE)')
        if line is None:
            return None, None
        if 'DID NOT CONVERGE' in line:
            return False, None
        if line:
            electron_info = self.get_electron_info()
            scf_energies = self.parse_scf_energies()
            conv_acc = self.parse_conv_acc()
            if electron_info is None or scf_energies is None or conv_acc is None:
                iteration = None
            else:
                iteration = {**scf_energies, **conv_acc, **electron_info}
            is_converged = self.check_for_until(re.compile(r'Self-consistency cycle converged'), re.compile(r'End self-consistency iteration'))
        return is_converged, iteration

    def get_scf_cycle(self, scf_cycle: Dict[str, Any]) -> Dict[str, Any]:
        while True:
            is_converged, iteration = self.get_iteration()
            if is_converged is None:
                break
            if iteration:
                scf_cycle['iterations'].append(iteration)
            if is_converged:
                scf_cycle['isConverged'] = True
                break
        return scf_cycle

    def get_scf_cycle_md_light(self) -> Dict[str, Any]:
        scf_cycle = {
            'structure': self.current_geometry,
            'iterations': [],
            'isConverged': False,
        }
        is_end = False
        iter = -1
        line = self.wait_for(re.compile(r'Convergence:'))
        if line is None:
            return None, None
        while not is_end:
            m = self.match_next_line(re.compile(self.float_regex))
            if m:
                iteration = {}
                if self.run_time_choices['spin'] == 'collinear' and len(m.groups()) >= 5:
                    iteration['convergenceAccuracy'] = {
                        'chargeDensityUp':  float(m.group(1)),
                        'chargeDensityDown':float(m.group(2)),
                        'eigenvalues':      float(m.group(3)),
                        'totalEnergy':      float(m.group(4)),
                    }
                    scf_cycle['iterations'].append(iteration)
                    iter += 1
                elif self.run_time_choices['spin'] == 'none' and len(m.groups()) >= 4:
                    iteration['convergenceAccuracy'] = {
                        'chargeDensity':    float(m.group(1)),
                        'eigenvalues':      float(m.group(2)),
                        'totalEnergy':      float(m.group(3)),
                    }
                    scf_cycle['iterations'].append(iteration)
                    iter += 1
                else:
                    is_end = True
            else:
                is_end = True

        scf_energies = self.scf_cycle['iterations'][iter]['scfEnergies'] = scf_energies['scfEnergies']
        electron_info = self.get_electron_info()
        scf_cycle['iterations'][iter]['electronInfo'] = electron_info['electronInfo']
        is_converged = self.check_for_until(re.compile(r'Self-consistency cycle converged'),
                                            re.compile(r'End self-consistency iteration|Reaching maximum number of scf iterations'))
        scf_cycle['isConverged'] = is_converged if is_converged is not None else False
        return scf_cycle

    def get_geo_from_lines(self, geo_lines: List[str]) -> Structure:
        geo = Structure()
        has_cartesian = False
        for line in geo_lines:
            if re.match(r'^[\s\t]*atom[\s\t]+', line):
                _, __, x, y, z, species = re.split(r'[\s\t]+', line)
                geo.add_atom_data([float(x), float(y), float(z)], species.strip())
                has_cartesian = True
            elif re.match(r'^[\s\t]*atom_frac[\s\t]+', line) and not has_cartesian:
                _, __, x, y, z, species = re.split(r'[\s\t]+', line)
                geo.add_atom_data([float(x), float(y), float(z)], species.strip(), True)
            elif re.match(r'^[\s\t]*lattice_vector[\s\t]+', line):
                if geo.lat_vectors is None:
                    geo.lat_vectors = []
                _, __, x, y, z = re.split(r'[\s\t]+', line)
                geo.lat_vectors.append([float(x), float(y), float(z)])
        return geo

    def get_initial_geometry(self) -> Structure:
        geo_lines = self.inputs['geometryIn']
        geo = self.get_geo_from_lines(geo_lines)
        self.system_information['nAtoms'] = len(geo.atoms)
        self.system_information['formulaUnit'] = self.get_formula_unit(geo)
        return geo

    @staticmethod
    def get_formula_unit(structure: Structure) -> str:
        species = {}
        for atom in structure.atoms:
            sp = atom['species']
            species[sp] = species.get(sp, 0) + 1
        formula_unit = ''.join(f"{key}{value if value > 1 else ''}" for key, value in species.items())
        return formula_unit

    def parse_updated_geo(self) -> Optional[Dict[str, Any]]:
        line = self.wait_for(re.compile(r'Geometry optimization: Attempting to predict improved coordinates\.'))
        if line is None:
            return None
        line = self.wait_for(re.compile(r'Maximum force component'))
        max_force_component = float(re.findall(self.float_regex, line)[0])

        line = self.wait_for(re.compile(r'Present geometry is'))
        if 'not yet converged.' in line:
            line = self.wait_for(re.compile(r'Updated atomic structure:'))
            if line is None:
                return None
            geo_lines = self.parse_until([re.compile(r'atom\s+'), re.compile(r'lattice_vector\s+')], re.compile(r'---------'))
            return {
                'force': max_force_component,
                'geo': self.get_geo_from_lines(geo_lines)
            }
        else:
            line = self.wait_for(re.compile(r'Final atomic structure:'))
            if line is None:
                return None
            return {
                'force': max_force_component,
                'geo': self.current_geometry
            }

    def get_mulliken_pop(self, soc: bool = False) -> Optional[Dict[str, List[float]]]:
        mulliken_charge = {}
        if soc:
            line = self.wait_for(re.compile(r'Performing scalar-relativistic Mulliken charge analysis'))
            if line is None:
                return None
            mulliken_charge['sr'] = self.read_mulliken()
            line = self.wait_for(re.compile(r'Performing spin-orbit-coupled Mulliken charge analysis'))
            if line is None:
                return None
            mulliken_charge['soc'] = self.read_mulliken()
        else:
            line = self.wait_for(re.compile(r'Performing Mulliken charge analysis on all atoms'))
            if line is None:
                return None
            mulliken_charge['simple'] = self.read_mulliken()
        return mulliken_charge

    def read_mulliken(self) -> List[float]:
        charge = []
        self.wait_for(re.compile(r'\| *atom'))
        for _ in range(self.system_information['nAtoms']):
            line = next(self.line_it)
            match = re.findall(self.float_regex, line)
            charge.append(float(match[1]))
        return charge

    def get_hirshfeld_pop(self) -> Optional[List[float]]:
        hirshfeld_charge = []
        self.wait_for(re.compile(r'Performing Hirshfeld analysis'))
        for _ in range(self.system_information['nAtoms']):
            line = self.wait_for(re.compile(r'Hirshfeld charge'))
            if line is None:
                return None
            hirshfeld_charge.append(float(re.findall(self.float_regex, line)[0]))
        return hirshfeld_charge

    def get_final_scf_energies_forces(self) -> Dict[str, Any]:
        final_energies = {}
        forces = []
        line = self.wait_for(re.compile(r'Energy and forces in a compact form'))
        if line is None:
            return {}
        line = next(self.line_it)
        final_energies['totalEnergy'] = float(re.findall(self.float_regex, line)[0])
        line = next(self.line_it)
        final_energies['totalEnergyCorrected'] = float(re.findall(self.float_regex, line)[0])
        line = next(self.line_it)
        final_energies['electronicFreeEnergy'] = float(re.findall(self.float_regex, line)[0])

        max_force_component = None
        if self.run_time_choices.get('has_forces', False):
            line = self.wait_for(re.compile(r'Total atomic forces'))
            for _ in range(self.system_information['nAtoms']):
                line = next(self.line_it)
                if line is None:
                    break
                force = [float(x) for x in re.findall(self.float_regex, line)]
                forces.append(force)
            max_force_component = max(max(abs(f) for f in force) for force in forces) if forces else None

        return {
            'finalScfEnergies': final_energies,
            'forces': forces,
            'maxForceComponent': max_force_component,
        }

    def normal_parser(self, file_text: str) -> None:
        """ This is the 'entry point' for parsing. """
        self.lines = file_text.split('\n')
        self.line_it = iter(self.lines)
        self.calculation_info = self.get_calculation_info()

        self.inputs['controlIn'] = self.get_input(
            re.compile(r'^\s+Parsing control\.in'),
            re.compile(r'^\s+Completed first pass over input file control\.in \.')
        )
        self.get_run_time_choices()
        self.inputs['geometryIn'] = self.get_input(
            re.compile(r'^\s+Parsing geometry\.in'),
            re.compile(r'^\s+Completed first pass over input file geometry\.in \.')
        )
        self.structure_in = self.get_initial_geometry()
        self.current_geometry = self.structure_in
        self.mulliken = None
        self.hirshfeld = None

        self.scf_loops = self.get_scf_loops()

        self.final_timings = self.get_final_timings()
        if self.final_timings is None:
            # not ideal but works at least
            self.line_it = iter(self.lines)
            self.final_timings = self.get_final_timings()
        self.memory = self.get_memory()
        if self.memory is None:
            self.line_it = iter(self.lines)
            self.memory = self.get_memory()
        self.exit_mode = self.get_exit_mode()

    def check_for_until(self, check_re: re.Pattern, stop_re: re.Pattern) -> Optional[bool]:
        for line in self.line_it:
            if re.search(check_re, line):
                return True
            elif re.search(stop_re, line):
                return False
        self.errors.append([f"Reached end of file before finding {stop_re.pattern}"])
        return None

    def get_scf_loops(self) -> List[Dict[str, Any]]:
        scf_loops = []
        while True:
            is_starting = self.check_for_until(
                re.compile(r'Begin self-consistency loop', re.IGNORECASE),
                re.compile(r'Leaving FHI-aims', re.IGNORECASE)
            )
            if is_starting is None:
                logger.info(f"Reached end of file after {len(scf_loops)} loops")
                break
            if is_starting:
                logger.info(f"Found start of SCF loop {len(scf_loops) + 1}")
                scf_cycle = {
                    'isConverged': False,
                    'iterations': [],
                    'structure': self.current_geometry,
                    'forces': [],
                    'maxForceComponent': None,
                    'finalScfEnergies': {}
                }
                output_level = self.run_time_choices.get('output_level', 'normal')
                if output_level == 'normal':
                    scf_cycle = self.get_scf_cycle(scf_cycle)
                elif output_level == 'MD_light':
                    scf_cycle = self.get_scf_cycle_md_light(scf_cycle)
                else:
                    logger.warning(f"Warning: Unknown output_level '{output_level}'. Using normal parsing.")
                    scf_cycle = self.get_scf_cycle(scf_cycle)

                logging.debug(f"SCF cycle converged: {scf_cycle['isConverged']}")

                final_scf_energies_forces = self.get_final_scf_energies_forces()
                scf_cycle['finalScfEnergies'] = final_scf_energies_forces.get('finalScfEnergies', {})
                scf_cycle['forces'] = final_scf_energies_forces.get('forces', [])
                scf_cycle['maxForceComponent'] = final_scf_energies_forces.get('maxForceComponent')

                scf_loops.append(scf_cycle)
            else:
                logger.info("Skipping non-SCF section")

        return scf_loops

    def get_input(self, start: re.Pattern, stop: re.Pattern) -> List[str]:
        c_lines = self.get_lines_in_between(start, stop)
        c_lines = c_lines[5:-2]  # remove the first six lines and the 2 last lines
        return c_lines

    def get_calculation_info(self) -> Dict[str, Dict[str, Any]]:
        calculation_info = {}
        calculation_info_reg_exs = [
            ('codeVersion', 'Code Version', re.compile(r'FHI-aims version\s+:\s+([0-9]+)')),
            ('commitNumber', 'Commit Number', re.compile(r'Commit number\s+:\s+([0-9,a-z]*)')),
            ('numberOfTasks', 'Number of Tasks', re.compile(r'Using\s+([0-9]*)\s+parallel tasks\.')),
        ]
        for line in self.line_it:
            if 'Obtaining array dimensions for all initial allocations:' in line:
                return calculation_info
            for key, info, regex in calculation_info_reg_exs:
                m = regex.search(line)
                if m:
                    calculation_info[key] = {'value': m.group(1), 'info': info}
        return None

    def get_final_timings(self) -> Optional[Dict[str, Dict[str, Any]]]:
        final_timings = {}
        # there are 11 instances of 'Total time' in this sample file
        # each has 2 floats (seconds) for max(cpu_time) and wall_clock(cpu1)
        # the Graphical interface uses wall_clock
        # TODO: fix this, currently it is not being parsed
        line = self.wait_for(re.compile(r' Total time  '))
        if line is None:
            return None
        final_timings['totalTime'] = {
            'value': re.findall(self.float_regex, line)[1],
            'info': 'Total Time'
        }
        return final_timings

    def get_memory(self) -> Optional[Dict[str, Dict[str, Any]]]:
        memory = {}
        line = self.wait_for(re.compile(r'   Maximum'))
        if line is None:
            return None
        m = re.findall(self.float_regex, line)
        memory['peakMemory'] = {
            'value': m[0],
            'info': 'Peak memory among tasks (MB)'
        }
        line_match = self.wait_for(re.compile(r'   Maximum'))
        if line_match is None:
            return memory
        m = re.findall(self.float_regex, line_match)
        memory['largestArray'] = {
            'value': m[0],
            'info': 'Largest tracked array allocation (MB)'
        }
        return memory

    def get_exit_mode(self) -> Optional[Dict[str, Any]]:
        exit_mode = {}
        line = self.wait_for(re.compile(r'Have a nice day\.'))
        if line is None:
            exit_mode['normalTermination'] = False
            exit_mode['errorMessage'] = 'Calculation did not terminate normally'
        else:
            exit_mode['normalTermination'] = True 
        # Check for specific error messages or warnings if needed
        # For example:
        # error_line = self.wait_for(re.compile(r'Error:'))
        # if error_line:
        #     exit_mode['errorMessage'] = error_line.strip()
        return exit_mode

    def scan_for_errors(self) -> List[str]:
        """ Searches for error messages found in the output file
            based on predefined regex patterns and adds them to self.errors

            This has been added to the original impl.
        """
        error_patterns = [
            r"BAD TERMINATION OF ONE OF YOUR APPLICATION PROCESSES",
            r"\*{3}\s*.*?\s*Error"
        ]
        combined_pattern = '|'.join(error_patterns)
        regex = re.compile(combined_pattern, re.IGNORECASE)
        for line in self.lines:
            if regex.search(line):
                self.errors_set.add(line.strip())



from typing import Dict, List, Optional, Any


class Output:
    def __init__(self):
        self.files: Dict[str, Dict[str, Any]] = {
            'dos': {},
            'absorption': {},
            'dielectric': {},
            'output': {},
            'input': {}
        }
        
        self.filled_up: bool = False
        self.inputs: Dict[str, Any] = {}
        self.results: List[Any] = []
        self.run_time_choices: Dict[str, Any] = {
            'calculationType': 'singlePoint',  # 'singlePoint', 'relaxation'
            'spin': 'none',  # 'none', 'collinear'
            'hasForces': False,
            'isPeriodic': False,
            'hasStress': False,
            'outputLevel': 'normal',  # 'normal', 'MD_light'
            'hasSOC': False,
            'hasHirshfeld': False,
            'hasMulliken': False,
            'hasMBD': False
        }
        self.system_information: Dict[str, Any] = {
            'nAtoms': 0,
            'formulaUnit': None,
            'volume': None
        }
        self.control_in: Optional[str] = None
        self.structure_in: Optional[Any] = None
        self.errors: List[str] = []
        self.ks_ev: Optional[Any] = None
        self.workflows: List[Any] = []

    def parse_files(self, files_data: List[Dict[str, Any]]) -> None:
        pass

    def is_filled_up(self) -> bool:
        return self.filled_up

    def is_relaxation(self) -> bool:
        return self.run_time_choices['calculationType'] == 'relaxation'

    def parse_output_file(self, file_name: Optional[str] = None) -> None:
        pass

    def parse_file(self, file_name: str, file_text: str) -> None:
        pass

    def get_structure(self) -> Optional[Any]:
        return self.structure_in

    def get_first_file_dos_data(self) -> Optional[Any]:
        return next(iter(self.files['dos'].values()), None)

    def get_bs_info(self) -> Any:
        # not all of these methods are even implemented
        pass



from typing import List, Optional, Set

from pydantic import BaseModel, Field


def dump(self, *args, **kwargs):
    return self.model_dump_json(*args, **kwargs)


BaseModel.dump = dump


class Results(BaseModel):
    total_energy:            float = Field(..., alias="Total Energy (eV)")
    fermi_energy:            float = Field(..., alias="Fermi Energy (eV)")
    highest_occupied_state:  float = Field(..., alias="Highest occupied state (eV)")
    lowest_unoccupied_state: float = Field(..., alias="Lowest unoccupied state (eV)")
    estimated_homo_lumo_gap: float = Field(..., alias="Estimated HOMO-LUMO gap (eV)")
    cell_volume:             float = Field(..., alias="Cell Volume (&#197;<sup>3</sup>)")


class CalculationSummary(BaseModel):
    code_version:                           str
    commit_number:                          str
    number_of_tasks:                        int
    total_time:                             float
    peak_memory_among_tasks_mb:             float
    largest_tracked_array_allocation_mb:    Optional[float]
    calculation_exited_regularly:           str


class DataSeries(BaseModel):
    data:  List[float]
    label: str  # Change of Eigenvalues, totalEnergy, chargeDensity...


class ChangeOfSumOfEigenvalues(BaseModel):
    charge_density:         Optional[DataSeries] = Field(None, alias="chargeDensity")
    charge_density_up:      Optional[DataSeries] = Field(None, alias="chargeDensityUp")
    charge_density_down:    Optional[DataSeries] = Field(None, alias="chargeDensityDown")
    eigen_values:           DataSeries = Field(..., alias="eigenvalues")
    total_energy:           DataSeries = Field(..., alias="totalEnergy")


class MaximumForceComponent(BaseModel):
    energy: DataSeries
    forces: DataSeries


class OutputData(BaseModel):
    """ GIMS OutputAnalyzer data model """
    results:                         Results
    calculation_summary:             CalculationSummary
    change_of_sum_of_eigenvalues:    List[ChangeOfSumOfEigenvalues]
    maximum_force_component:         Optional[MaximumForceComponent]



"""
Application configuration file

Author: Iker Hurtado
"""

from typing import Dict, List, Any, Optional


class Conf:
    """Object storing the configuration"""

    # Default settings
    settings: Dict[str, Any] = {
        "decimalDigits": 6,
        "symmetryThreshold": 1e-5,
    }

    BASE_FOLDER: str = ''  # 'static/'

    # Species codes
    ELEMENT_NAMES: Dict[str, int] = {
        element: i for i, element in enumerate([
            'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si',
            'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni',
            'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb',
            'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe',
            'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho',
            'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',
            'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np',
            'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Ha', 'Sg',
            'Ns', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og'
        ])
    }

    # StructureViewer constants
    RADIUS_FACTOR: float = 0.5
    MISSING_RADIUS: float = 0.2

    ELEMENT_RADII: List[float] = [
        0.31, 0.28, 1.28, 0.96, 0.84, 0.76, 0.71, 0.66, 0.57, 0.58,
        1.66, 1.41, 1.21, 1.11, 1.07, 1.05, 1.02, 1.06, 2.03, 1.76, 1.7, 1.6, 1.53,
        1.39, 1.39, 1.32, 1.26, 1.24, 1.32, 1.22, 1.22, 1.2, 1.19, 1.2, 1.2, 1.16,
        2.2, 1.95, 1.9, 1.75, 1.64, 1.54, 1.47, 1.46, 1.42, 1.39, 1.45, 1.44, 1.42,
        1.39, 1.39, 1.38, 1.39, 1.4, 2.44, 2.15, 2.07, 2.04, 2.03, 2.01, 1.99, 1.98,
        1.98, 1.96, 1.94, 1.92, 1.92, 1.89, 1.9, 1.87, 1.87, 1.75, 1.7, 1.62, 1.51,
        1.44, 1.41, 1.36, 1.36, 1.32, 1.45, 1.46, 1.48, 1.4, 1.5, 1.5, 2.6, 2.21,
        2.15, 2.06, 2.0, 1.96, 1.9, 1.87, 1.8, 1.69, MISSING_RADIUS, MISSING_RADIUS,
        MISSING_RADIUS, MISSING_RADIUS, MISSING_RADIUS, MISSING_RADIUS, MISSING_RADIUS
    ]

    MISSING_COLOR: str = '#000000'

    ELEMENT_COLORS: List[str] = [
        '#ffffff', '#d9ffff', '#cc80ff', '#c2ff00', '#ffb5b5', '#909090', '#3050f8', 
        '#ff0d0d', '#90e050', '#b3e3f5', '#ab5cf2', '#8aff00', '#bfa6a6', '#f0c8a0', '#ff8000', 
        '#ffff30', '#1ff01f', '#80d1e3', '#8f40d4', '#3dff00', '#e6e6e6', '#bfc2c7', '#a6a6ab', 
        '#8a99c7', '#9c7ac7', '#e06633', '#f090a0', '#50d050', '#c88033', '#7d80b0', '#c28f8f', 
        '#668f8f', '#bd80e3', '#ffa100', '#a62929', '#5cb8d1', '#702eb0', '#00ff00', '#94ffff', 
        '#94e0e0', '#73c2c9', '#54b5b5', '#3b9e9e', '#248f8f', '#0a7d8c', '#006985', '#c0c0c0', 
        '#ffd98f', '#a67573', '#668080', '#9e63b5', '#d47a00', '#940094', '#429eb0', '#57178f', 
        '#00c900', '#70d4ff', '#ffffc7', '#d9ffc7', '#c7ffc7', '#a3ffc7', '#8fffc7', '#61ffc7', 
        '#45ffc7', '#30ffc7', '#1fffc7', '#00ff9c', '#00e675', '#00d452', '#00bf38', '#00ab24', 
        '#4dc2ff', '#4da6ff', '#2194d6', '#267dab', '#266696', '#175487', '#d0d0e0', '#ffd123', 
        '#b8b8d0', '#a6544d', '#575961', '#9e4fb5', '#ab5c00', '#754f45', '#428296', '#420066', 
        '#007d00', '#70abfa', '#00baff', '#00a1ff', '#008fff', '#0080ff', '#006bff', '#545cf2', 
        '#785ce3', '#8a4fe3', '#a136d4', '#b31fd4', '#b31fba', '#b30da6', '#bd0d87', '#c70066', 
        '#cc0059', '#d1004f', '#d90045', '#e00038', '#e6002e', '#eb0026'
    ]

    @staticmethod
    def get_species_color(species: Optional[str]) -> str:
        """
        Returns the species color code
        :param species: Element symbol
        :return: Color code as string
        """
        if species is None:
            return Conf.MISSING_COLOR
        return Conf.ELEMENT_COLORS[Conf.ELEMENT_NAMES[species]]

    @staticmethod
    def set_species_color(species: str, color: str) -> None:
        """
        Sets a species color code overwriting the original one
        :param species: Element symbol
        :param color: New color code
        """
        print(f'setSpeciesColor: {species}, {Conf.ELEMENT_NAMES[species]}, {Conf.ELEMENT_COLORS[Conf.ELEMENT_NAMES[species]]}, {color}')
        Conf.ELEMENT_COLORS[Conf.ELEMENT_NAMES[species]] = color

    @staticmethod
    def get_species_radius(species: Optional[str]) -> float:
        """
        Returns the species radius
        :param species: Element symbol
        :return: Radius as float
        """
        if species is None:
            return Conf.MISSING_RADIUS
        return Conf.ELEMENT_RADII[Conf.ELEMENT_NAMES[species]]

    @staticmethod
    def get_radius_and_color(species: Optional[str]) -> Dict[str, Any]:
        """
        Returns the species radius and color
        :param species: Element symbol
        :return: Dictionary with radius and color
        """
        if species is None:
            return {"radius": 0.5, "color": Conf.MISSING_COLOR}
        index = Conf.ELEMENT_NAMES[species]
        return {
            "radius": Conf.ELEMENT_RADII[index] * Conf.RADIUS_FACTOR,
            "color": Conf.ELEMENT_COLORS[index]
        }

    @staticmethod
    def get_species_atomic_number(species: str) -> int:
        """
        Returns the species atomic number
        :param species: Element symbol
        :return: Atomic number as integer
        """
        return Conf.ELEMENT_NAMES[species] + 1



from ataims.version import __version__
from ataims.parser import parse_outputfile
from ataims.exceptions import FHIOutputParserError


__all__ = ['parse_outputfile', 'FHIOutputParserError', '__version__']



import numpy as np

# import output_parser.UserMsgBox as UserMsgBox
from .mathlib import matrix_dot, tensor_dot, add_arrays, invert33, minmax, determinant33
from .structure import Structure
# from .conf import Conf


def get_tokenized_lines(text):
    """
    Divides a text into lines and returns them in a list
    :param text: str
    :return: list of str
    """
    return text.replace(r'[ \t]+', ' ').split('\n')

# rewrite
# async def load_data_file(file_name, file_type, handler):
#     """
#     Loads a file from the application data folder and passes the content to a handler
#     :param file_name: str
#     :param file_type: str
#     :param handler: function
#     """
#     data = None
#     try:
#         response = requests.get(Conf.BASE_FOLDER + 'data/' + file_name)
#         if file_type == 'text':
#             data = response.text
#     except Exception as err:
#         print('loadDataFile error:', err)
#     handler(data)


def generate_supercell(dim_array, source_struct):
    """
    Generates a supercell of the given structure.
    :param dim_array: list of str
    :param source_struct: Structure
    :return: Structure, new structure representing the supercell
    """
    supercell_matrix = np.zeros((3, 3), dtype=int)
    if len(dim_array) == 9:
        for i in range(len(dim_array)):
            supercell_matrix[i // 3, i % 3] = int(dim_array[i])
    elif len(dim_array) == 3:
        for i in range(len(dim_array)):
            supercell_matrix[i % 3, i % 3] = int(dim_array[i])

    lattice = source_struct.lat_vectors
    super_lattice = matrix_dot(supercell_matrix, lattice, np.zeros_like(lattice))
    frac_points = lattice_points_in_supercell(supercell_matrix)

    new_struct = Structure()
    new_struct.file_source = source_struct.file_source + " (supercell)"
    new_struct.update_lattice_vectors(super_lattice, True)

    points = matrix_dot(frac_points, super_lattice, np.zeros_like(frac_points))
    for p in points:
        for atom in source_struct.atoms:
            new_cart = add_arrays(atom['position'], p)
            new_fract = [x + 1 if x < 0 else x - 1 if x >= 1 else x for x in new_struct.get_fractional_coordinates(new_cart)]
            new_position = new_struct.get_cartesian_coordinates(new_fract)
            new_struct.add_atom_data(new_position, atom['species'], False, atom['initMoment'], atom['constraint'], atom['charge'])
    return new_struct


def lattice_points_in_supercell(supercell_matrix):
    """
    Returns all lattice points of the old unit cell within the new supercell.
    :param supercell_matrix: list of lists
    :return: list of lists
    """
    diags = [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0], [0, 1, 1], [1, 0, 1], [1, 1, 0], [1, 1, 1]]
    d_points = np.zeros_like(diags)

    matrix_dot(diags, supercell_matrix, d_points)
    mins = minmax(d_points, 0, np.min)
    maxs = [x + 1 for x in minmax(d_points, 0, np.max)]

    ranges = [np.arange(mins[i], maxs[i]) for i in range(3)]

    ar = tensor_dot(ranges[0], [[1, 0, 0]])
    br = tensor_dot(ranges[1], [[0, 1, 0]])
    cr = tensor_dot(ranges[2], [[0, 0, 1]])

    all_points = [add_arrays(ar[i], add_arrays(br[j], cr[k])) for i in range(len(ar)) for j in range(len(br)) for k in range(len(cr))]
    frac_points = matrix_dot(all_points, invert33(supercell_matrix), np.zeros_like(all_points))

    result = [p for p in frac_points if all(-1e-10 < x < 1 - 1e-10 for x in p)]
    assert len(result) == determinant33(supercell_matrix), 'We are missing some lattice points. Check precision of supercell matrix'
    return result

# async def generate_slab(slab_data, structure):
#     """
#     Generates a slab on the backend
#     :param slab_data: dict
#     :param structure: Structure
#     :return: Structure, slab structure
#     """
#     if not structure.atoms:
#         return
#     send_structure = {
#         'cell': structure.lat_vectors,
#         'positions': structure.atoms,
#         'slab_data': slab_data,
#         'fileName': structure.file_source
#     }
#     end_point = Conf.BASE_URL + ('/terminate-slab' if 'terminations' in slab_data else '/get-slab')
#     response = requests.post(end_point, headers={'Content-Type': 'application/json;charset=utf-8'}, json=send_structure)

#     if not response.ok:
#         UserMsgBox.setError('Unknown server ERROR')
#         return

#     slab_json = response.json()
#     slab_structure = get_structure_from_json(slab_json)
#     slab_structure.extraData['terminations'] = slab_json['terminations']
#     return slab_structure

# async def get_structure_info(structure):
#     """
#     Gets structure information from the backend
#     :param structure: Structure
#     :return: dict, JSON with structure information
#     """
#     structure_data = None
#     if structure.atoms:
#         send_structure = {
#             'cell': structure.lat_vectors,
#             'positions': structure.atoms,
#             'symThresh': Conf.settings['symmetryThreshold']
#         }
#         response = requests.post(Conf.BASE_URL + '/update-structure-info', headers={'Content-Type': 'application/json;charset=utf-8'}, json=send_structure)
#         if not response.ok:
#             UserMsgBox.setError('Unknown server ERROR')
#             raise Exception(f"HTTP error: {response.status}")
#         structure_data = response.json()
#     return structure_data.get('structureInfo') if structure_data else None


def get_structure_from_json(json_data):
    """
    Returns a Structure from a JSON object
    :param json_data: dict
    :return: Structure
    """
    structure = Structure()
    structure.file_source = json_data['fileName']
    structure.structure_info = json_data['structureInfo']
    structure.lat_vectors = json_data['lattice'] if json_data['lattice'] is not None else None
    for atom in json_data['atoms']:
        structure.add_atom_data(atom[0], atom[1], False, atom[3], atom[4], atom[5])  # cartesian coordinates
    return structure


def get_structure_from_file_content(file_name, file_content):
    """
    Returns a Structure from text content in a file (FHIaims geometry file)
    :param file_name: str
    :param file_content: str
    :return: Structure
    """
    file_ext = file_name.split('.')[-1]
    structure = parse_geometry_in_file_format(file_content) if file_ext == 'in' else None
    structure.file_source = file_name
    return structure


def parse_geometry_in_file_format(text):
    """
    Parses the FHIaims geometry file format and returns a Structure.
    :param text: str
    :return: Structure
    """
    VECTOR_KEYWORD = 'lattice_vector'
    ATOM_CART_KEYWORD = 'atom'
    ATOM_FRAC_KEYWORD = 'atom_frac'
    ATOM_INIT_MOMENT_KEYWORD = 'initial_moment'
    ATOM_INIT_CHARGE_KEYWORD = 'initial_charge'

    lines = get_tokenized_lines(text)
    structure = Structure()
    structure.lat_vectors = [] if VECTOR_KEYWORD in text else None

    for line in lines:
        tokens = line.strip().split(' ')
        if tokens[0] == VECTOR_KEYWORD:
            structure.lat_vectors.append([float(tokens[1]), float(tokens[2]), float(tokens[3])])
        elif tokens[0] in {ATOM_CART_KEYWORD, ATOM_FRAC_KEYWORD}:
            atom_pos = [float(tokens[1]), float(tokens[2]), float(tokens[3])]
            structure.add_atom_data(atom_pos, tokens[4], tokens[0] == ATOM_FRAC_KEYWORD)
        elif tokens[0] == ATOM_INIT_MOMENT_KEYWORD:
            structure.set_last_atom_init_moment(float(tokens[1]))
        elif tokens[0] == ATOM_INIT_CHARGE_KEYWORD:
            structure.set_last_atom_charge(float(tokens[1]))

    return structure


def get_geometry_in_texts(structure):
    # Implementation needed
    pass


def get_input_xml_texts(structure):
    # Implementation needed
    pass


def get_input_text_files_url(structure, fhi_aims=True):
    # Implementation needed
    pass


def get_geometry_files_testing(structure, fhi_aims=True):
    # Implementation needed
    pass


def deep_clone(obj):
    from copy import deepcopy
    return deepcopy(obj)


def int_validator(value):
    try:
        return int(value)
    except ValueError:
        return ''


def int_input_validator(value):
    if value in ["0", "-"]:
        return value
    try:
        return int(value)
    except ValueError:
        return ''


def float_validator(value):
    try:
        return float(value)
    except ValueError:
        return ''


def float_input_validator(value):
    if value in [".", "0", "-"]:
        return value
    try:
        return float(value)
    except ValueError:
        return ''



from typing import List, Dict, Optional, Union, Any
import numpy as np

from ataims.mathlib import multiply_scalar, add_arrays
from ataims.conf import Conf


def solve33(lattice_vectors: List[List[float]], cartesian_coords: List[float]) -> List[float]:
    """
    Solves the linear system to get the fractional coordinates from Cartesian coordinates
    :param lattice_vectors: 3x3 list of lattice vectors
    :param cartesian_coords: 1x3 list of Cartesian coordinates
    :return: 1x3 list of fractional coordinates
    """
    lattice_matrix = np.array(lattice_vectors)
    cartesian_coords = np.array(cartesian_coords)
    # Solve for fractional coordinates: lattice_matrix * fractional = cartesian_coords
    fractional_coords = np.linalg.solve(lattice_matrix.T, cartesian_coords)
    return fractional_coords.tolist()


class Structure:
    """
    A class modeling an atomic structure, either a periodic or non-periodic system.
    It's usually gotten from a file.
    """

    def __init__(self) -> None:
        self.lat_vectors: Optional[List[List[float]]] = None  # List of lists (3x3)
        self.atoms: List[Dict[str, Union[List[float], str, float, bool]]] = []
        self.file_source: str = ""
        self.structure_info: Dict[str, Any] = {}
        self.extra_data: Dict[str, Any] = {}
        self.id: Optional[str] = None

    def reset(self) -> None:
        """Reset the structure"""
        self.lat_vectors = None
        self.atoms = []
        self.file_source = None
        self.structure_info = None
        self.extra_data = None

    def is_a_periodic_system(self) -> bool:
        """Returns if it's a periodic system or not"""
        return self.lat_vectors is not None

    def remove_lattice_vectors(self) -> None:
        """Removes the lattice vectors becoming a non-periodic system"""
        self.lat_vectors = None

    def update_lattice_vectors(self, vectors: List[List[float]], scale_atoms_position: bool) -> None:
        """
        Updates the lattice vectors and scales the atoms positions if specified
        :param vectors: List of lists
        :param scale_atoms_position: bool
        """
        fract_positions = []
        if scale_atoms_position:
            fract_positions = [self.get_fractional_coordinates(atom['position']) for atom in self.atoms]

        self.lat_vectors = vectors

        if scale_atoms_position:
            for i, fract_pos in enumerate(fract_positions):
                self.atoms[i]['position'] = self.get_cartesian_coordinates(fract_pos)

    def create_atom(self, cart_coords: List[float], species: str, moment: float, constraint: bool, charge: float) -> Dict[str, Union[List[float], str, float, bool]]:
        """
        Creates a new atom object (it's not added to the structure)
        :return: dict
        """
        return {
            'position': cart_coords,
            'species': species,
            'initMoment': moment,
            'constraint': constraint,
            'charge': charge,
            'radius': Conf.get_species_radius(species)
        }

    def add_atom_data(self, pos: List[float], species: str, is_fractional: bool = False, moment: float = 0, constraint: bool = False, charge: float = 0, index: Optional[int] = None) -> None:
        """
        Creates and add a new atom object
        """
        c_coors = self.get_cartesian_coordinates(pos) if is_fractional else pos
        atom = self.create_atom(c_coors, species, moment, constraint, charge)
        if index is None:
            self.atoms.append(atom)
        else:
            self.atoms.insert(index, atom)

    def add_undefined_atom(self) -> int:
        """
        Creates an undefined atom.
        :return: int, new atom position in the structure
        """
        self.add_atom_data([0, 0, 0], None)
        return len(self.atoms) - 1

    def update_atom_species(self, atom_index: int, species: str) -> None:
        """Updates a structure atom species"""
        self.atoms[atom_index]['species'] = species

    def update_atom_init_moment(self, atom_index: int, init_moment: float) -> None:
        """Updates a structure atom initialization moment"""
        self.atoms[atom_index]['initMoment'] = init_moment

    def update_atom_constraint(self, atom_index: int, constraint: bool) -> None:
        """Updates a structure atom constrain"""
        self.atoms[atom_index]['constraint'] = constraint

    def update_atom_charge(self, atom_index: int, charge: float) -> None:
        """Updates a structure atom charge"""
        self.atoms[atom_index]['charge'] = charge

    def update_atom_position(self, atom_index: int, values: List[float], is_fract: bool = False) -> None:
        """Updates a structure atom position"""
        for i in range(3):
            self.update_atom_coor_value(atom_index, i, values[i], is_fract)

    def update_atom_coor_value(self, atom_index: int, coor_index: int, value: float, is_fract: bool = False) -> None:
        """Updates a coordinate of a structure atom"""
        if atom_index >= len(self.atoms):
            atom = self.create_atom([0, 0, 0], None, 0, False, 0)
            self.atoms.append(atom)  
        if is_fract:
            self.atoms[atom_index]['position'][coor_index] = self.get_cartesian_coordinates([value if i == coor_index else 0 for i in range(3)])[coor_index]
        else:
            self.atoms[atom_index]['position'][coor_index] = value

    async def update_info(self) -> None:
        """Updates information on this structure"""
        # looks like network requests
        from util import get_structure_info
        self.structure_info = await get_structure_info(self)

    def set_last_atom_init_moment(self, moment: float) -> None:
        """Sets the initialization moment to the last atom"""
        self.atoms[-1]['initMoment'] = moment

    def set_last_atom_charge(self, charge: float) -> None:
        """Sets the initialization charge to the last atom"""
        self.atoms[-1]['charge'] = charge

    def get_cartesian_coordinates(self, fract_coors: List[float]) -> List[float]:
        """
        Returns the cartesian coordinates corresponding to the
        fractional coordinates passed as the argument.
        """
        if not self.is_a_periodic_system():
            return fract_coors
        cart_pos = multiply_scalar(self.lat_vectors[0], fract_coors[0])
        cart_pos = add_arrays(cart_pos, multiply_scalar(self.lat_vectors[1], fract_coors[1]))
        cart_pos = add_arrays(cart_pos, multiply_scalar(self.lat_vectors[2], fract_coors[2]))
        return cart_pos

    def is_magnetic(self) -> bool:
        """
        Finds if the structure is magnetic, that is, if any of initial moments has been set
        :return: bool
        """
        return any(atom['initMoment'] != 0 for atom in self.atoms)

    def get_atom_fract_position(self, i: int) -> Optional[List[float]]:
        """Returns a structure atom fractional coordinates"""
        return self.get_fractional_coordinates(self.atoms[i]['position'])

    def get_fractional_coordinates(self, cart_coors: List[float]) -> Optional[List[float]]:
        """
        Returns the fractional coordinates corresponding to the
        cartesian coordinates passed as the argument.
        """
        return solve33(self.lat_vectors, cart_coors) if self.is_a_periodic_system() else None



import sys
import argparse
import logging

from ataims import parse_outputfile


def main():
    """Basic CLI. For programmatic use `from ataims import parse_outputfile` ."""
    parser = argparse.ArgumentParser(
                    prog='ataims',
                    description='Parses FHI-aims, Exciting and Quantum Espresso outputs',
    )
    parser.add_argument('filename', type=str, help='Path to the output file')
    parser.add_argument('-v', '--verbose', action='store_true')
    args = parser.parse_args()
    fname = args.filename
    if args.verbose:
        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    else:
        logging.basicConfig(level=logging.WARNING, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s') 
    sys.stdout.write(parse_outputfile(fname, as_set=False).model_dump_json() + '\n')


if __name__ == '__main__':
    main()
